#define SLOW_MPI 0                               /* old version, obsolete */
#define MAX_out 1000        /* maximum lines for debugging/verbose output */
#define UseAlpha      0     /* 1: parameter alpha used, 0: else (default) */
//#define StartAlpha    0.0                /* start angle   */
//#define StartAlpha    (M_PI/6.0)                /* start angle triangle AF in external field 1  */
#define StartAlpha    (M_PI/2.0)                /* start angle triangle AF in external field 2  */
#define StepAlpha     (0.001*M_PI)                         /* step of alpha */
#define StopAlpha     (StartAlpha+2*M_PI)                /* stop angle    */
#define SearchAlphaMin 0   /* 1: search alpha for minimal energy, 0: else */
//#define USR_FILE  "sample_bipartite_external_field.cpp"          /* alpha dependency of terms */ 
//#define USR_FILE  "sample_triangle_external_field_1.cpp"          /* alpha dependency of terms */ 
#define USR_FILE  "sample_triangle_external_field_2.cpp"          /* alpha dependency of terms */ 
// set UseExpAlpha equal to 0 if you don't want to use user defined file in find expectation values.
#define UseExpAlpha 0
//#define USR_EXP_FILE  "sample_bipartite_external_field_expectation.cpp" /* alpha dependency of terms: expectation values */
//#define USR_EXP_FILE  "sample_triangle_external_field_expectation_1.cpp" /* alpha dependency of terms: expectation values */
#define USR_EXP_FILE  "sample_triangle_external_field_expectation_2.cpp" /* alpha dependency of terms: expectation values */
#define SearchBetaMin 0  /* 1: search beta for minimal energy, 0: else */
#define SearchGammaMin 0   /* 1: search beta for minimal energy, 0: else */
/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:      CCM (solve.C)
  
  Written by: Damian J. J. Farnell

  Date:      Wed Mar 15 15:20:28 GMT 2000

  Updated:   Joerg Schulenburg 2004-2008

  ===================================================

  ---------------------------------------------------

  PROGRAM NAME: SOLVE.C 

  PURPOSE: Routine that solves saved CCM
           equations using newton-raphson
	   or direct iteration. Sublattice
           magnetisation determined and 
	   ground-state energy (+ derivatives).

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"
#include "nr.h"
#include "nrutil.h"
#include "storage.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

char gs_name[100]="";
char input_name[200];
// char configurations[200];

/* Parameter array of size Params.bonds d1*H1+d2*H2+... */
// double (*delta)[62+1]=NULL;
// double (*dincr)[62+1]=NULL; // increment to next delta (derivation)
double delta[5][62+1]; // depends from usr_angle*usr_parameter=f(bond,term63)
double dincr[5][62+1]; // increment to next delta (derivation)

/* -------------------------------------------------------------
 *  array of parameters depending on one variable parameter   
 *  sample: d1=cos(d), d2=sin(d) with H= d1*H1+d2*H2         
 */
static int PFLINE= __LINE__ + 1;   // this is only a localizator
// depends from alpha only? can it depend from term (term13 or term63)?
double usr_parameter(int bonds, int term, double alpha)
{
    /* add your model specifique parameter here! */
    //  tdelta * H1 + H2 + H3 + ...
//    if (bonds==1) return alpha;
    if (bonds==1 || bonds==2) return alpha; // condition for triangular AF in external field: both model states

    //  cos(alpha) * H1 + sin(alpha) * H2 + H3 + ...
 // if (bonds==1) return cos(alpha*M_PI);
 // if (bonds==2) return sin(alpha*M_PI); 
 // if (bonds==3) return 1; 

    return 1.0;
}

#if UseAlpha
// include user defined function usr_angle(bond,term,alpha,beta,gamma)
#include USR_FILE  
#endif

#if UseExpAlpha
#include USR_EXP_FILE  
#endif

double findEnergy(double *x);
void sublattice_term(int length, int (*configuration)[5], int *subTerm);
double findStagMag(double *x, double *bra_x, double *subLatticeMag);
void findBraState(double *x, double *bra_x);
void findBraStateRHS(double *rhs, double *x);
void findBraEquations(double **m, double *x);
int directIteration(int maximumNumIterations, int numConfigs, 
		    double tolerance, int &iterations, double *oldX);
int directIterationBraState(int maximumNumIterations, int numConfigs, 
			    double tolerance, int &iterations, 
			    double *oldX, double *oldBraX);


void solve(void)
{
  int i, j, k;
  double *x, *bra_x;
  // char root_name[200], append[]=".dat";
  char output[200]; // , output2[200];
  //  char path[]="../Data/";
  //  char v_j2[]="_v_j2";
  //  int  **configuration, length;
  int iterations, iterations2;
  double energy=0;
  //  double energy2;
  //  double specHeatA, specHeatC;
  int status; // , stat;
  int sign = 1;

  double stagMag;
  /* we need it to find the inflection points of E(p) curves */
  double Energy1, Energy2;
  double dEnergy, ddEnergy;
  double stagMag1, stagMag2;
  double dStagMag, ddStagMag;
  double alpha=0.0;
  double beta=0.0;
  double gamma=0.0;
  
  double start, stop, tdelta;
#if UseAlpha
  double d_alpha, old_Energy, minE=1e9, a_minE=0;
  double d_beta, d_gamma,old_beta_Energy,old_gamma_Energy;
  double step_d_beta=0.1,step_d_gamma=0.1;
#endif

  double subLatticeMag[MAX_UNIT_CELL+3]={0.0};

  double *g_exp;
  g_exp=      dvector(0,Params.nExpect+5);

  // configuration=imatrix(0,MAX_LSUBM+1,0,5);

  time_t theTime, t0, t1;

  //  solutionStrategy=1;
  //  momentum=0.7;         /* see Constants.h */

  if (PARALLEL_FLAG > 0 && solutionStrategy == 1){
    printf("Direct Iteration only for parallel processing...\n"
	   "Exiting!!\n\n");
    ERR(0);
  }

  time(&theTime);
  srand((int) theTime);

   strcpy(input_name,    Control.name); strcat(input_name,    ".terms");
   strcpy(output,        Control.name); strcat(output,        ".output");
// strcpy(configurations,Control.name); strcat(configurations,".configs");
   strcpy(gs_name,       Control.name); strcat(gs_name,       "_gs");

  if (myid==0) printf(" solve: myid=%d input=%s output=%s gs=%s\n",
         myid,input_name,output,gs_name);

  // terms_fp=fopen(input_name,"r"); => opened in ccm.C now

  if (myid == 0) {
    fp2=fopen(output,"w");  
    if (!fp2) {
      printf("Cannot Open %s...\nExiting!!\n\n", output);
      exit(1);
    }
  }

  // ground-state energy equation.

  // fp_gs=fopen(gs_name,"r");
  
  Energy2=1.0;    Energy1=2.0;   /* set initial values, no meaning */
  stagMag2=1.0;   stagMag1=2.0;

  x=      dvector(0,NumberConfigurations+5);
  bra_x=  dvector(0,NumberConfigurations+5);

  for(i=1;i<=NumberConfigurations;i++) { 
    x[i]=    0.000001; 
    bra_x[i]=0.000000001; 
  }
  // x[1]=0.1; // testwise, ToDo: failes, why?
  x[0]=    1.0; /* dummy index set to 1.0 for simplification, Joerg Sep05 */
  bra_x[0]=1.0; /* dummy index set to 1.0 for simplification, Joerg Sep05 */

  // x[2] = 0.064048;
  // x[3] = 0.000000;
  // x[4] = 0.000000;
  // x[5] = 0.064048;
  // x[6] = 0.064048;
 
  if (Control.output > -1 && myid == 0) {
    printf("============================================================\n\n");
    printf("SOLVE CCM EQUATIONS: START ALGORITHM.\n\n");
    printf("============================================================\n\n");
  }

  if (increment > 0.0) { start= minimum;  stop= maximum; }
    else               { start=-maximum;  stop=-minimum; }

   // Joerg Sep05: changing parameterfunction gives you more flexibility
  if (myid==0){ 
    printf("usr_parameter can be defined in "__FILE__" L%d\n",PFLINE);
    printf("solutionStrategy= %d (%s)\n",
            solutionStrategy,
          ((solutionStrategy==1)?"newton-raphson":"direct iteration"));
  }
  /* array of parameters depending on one variable Parameter
   *  sample: d1=cos(d), d2=sin(d) with H= d1*H1+d2*H2
   */ 
  // delta = malloc((Params.bonds+1)*sizeof(double)*63);
  // dincr = malloc((Params.bonds+1)*sizeof(double)*63);
  // if (!delta || !dincr) {
  //   printf("input.C: delta = malloc() failed...\nExit!\n");
  //   exit(1);
  // }
  
  sign = ((increment<0.0) ? -1 : +1);

#if UseAlpha
    alpha =  StartAlpha;  // starting angle, 
    d_alpha = StepAlpha;  // starting angle step
    beta=-0.5*M_PI;
    gamma=0.5*M_PI;
#if SearchGammaMin
   step_d_gamma = StepAlpha;  // starting angle step
#endif
#if SearchBetaMin
   step_d_beta = StepAlpha;  // starting angle start
#endif
#endif

  // loop over different couplings J, go from known to unknown angle
  for(tdelta=start; tdelta<=stop; tdelta+=fabs(increment)) {

    // D1(Delta)*H1 + D2(Delta)*H2 + D3(Delta)*H3 + ... , Joerg Oct05
    if (myid==0)
    printf("\nModel Parameter Delta = %6.3f\n", sign*tdelta);

    for (int i=1; i<= Params.bonds; i++) {
      int ja=1; /* start range */ 
      for (j=1; j<=62; j++) {  // ToDo: use term_13[63]
        delta[i][j] = usr_parameter(i,j,sign* tdelta );
        /* Joerg: dincr is only used for precondition (shift startvector) */
        dincr[i][j] = usr_parameter(i,j,sign*(tdelta+0.5*fabs(increment)))
                    - usr_parameter(i,j,sign*(tdelta-0.5*fabs(increment)));
        if (j>1 && delta[i][ja]!=delta[i][j] && myid==0) {
          printf(" Delta[bond=%d][term=%d...%d]= %6.3f\n",i,ja,j-1,delta[i][ja]);
          ja=j; /* next range */
        }
      }
      if (myid==0)
      printf(" Delta[bond=%d][term=%d...%d]= %6.3f\n",i,ja,j-1,delta[i][ja]);
    }

#if UseAlpha
    old_Energy = 1.0e21; // biggest possible value
    // find the the minimum E(angle), we assume that exact one minimum exist
    // increase accuracy of angle from 1e-1 to 1e-9 (interval)
    // ToDo: that seems not a good algorithm, use Newton method at least
    // go down with const steps, if it goes up, return with smaller steps
    for (d_alpha=StepAlpha; fabs(d_alpha)> 2.0e-9; alpha += d_alpha) {

	  old_beta_Energy=1.0e21;
       for (d_beta=fabs(d_alpha); fabs(d_beta)> fabs(d_alpha)/4.0; beta += d_beta) {
	   old_gamma_Energy=1.0e21;
       for (d_gamma=fabs(d_alpha); fabs(d_gamma)> fabs(d_alpha)/4.0; gamma += d_gamma) {
 
       // find the best angle (lowest energy) = angle +/- d_angle
       //
       // E                            5       6       7
       // step ------------------------1-------3---4-562------------- 
       // angle Pi/2                 +0.1   +0.125   +0.15

       for (int i=1; i<= Params.bonds; i++)
         for (j=1; j<=62; j++) {
           delta[i][j] = usr_angle(i,j,alpha,beta,gamma )
                       * usr_parameter(i,j,sign* tdelta );
           dincr[i][j] = usr_parameter(i,j,sign*(tdelta+0.5*fabs(increment)))
                       - usr_parameter(i,j,sign*(tdelta-0.5*fabs(increment)));
           dincr[i][j]*=usr_angle(i,j,alpha,beta,gamma );
           // dincr[i][j] = ??? 
         }
#endif
    iterations=0;
    iterations2=0;
    time(&t0);      // start time
    
    if (solutionStrategy==1) {
      
     // (a) newton-raphson ket state + LU decomp. bra state.

      status=mnewt((int) MAX_NUM_ITERATIONS, NumberConfigurations, 
		   (double) TOLERANCE, (double) TOLERANCE, iterations, x);
      
      // findBraState(x, bra_x);

    } else {
      
      // (b) direct iterations for both ket and bra states.

      status=directIteration((int) MAX_NUM_ITERATIONS,
             NumberConfigurations, (double) TOLERANCE, 
	     iterations, x);

      // status=directIterationBraState((int) MAX_NUM_ITERATIONS, 
	//     NumberConfigurations, (double) TOLERANCE, 
	//     iterations2, x, bra_x); 
        // status == num_iterations
    }
    time(&t1);

    if (myid==0 && status < 0)
      printf("status= %d\n", status);

    if (iterations >= MAX_NUM_ITERATIONS) {
       printf("WARNING: iterations=%d >= MAX_NUM_ITERATIONS\n", iterations);
       continue; // Mar08
       // break;
    }
    
    energy=findEnergy(x);
    if (Control.output>0 && myid==0) // print out later?
        printf("Eg= %18.14f it= %4d  Eg.i100t= %6.2fm\n",
           energy, iterations, (t1-t0)/.6/iterations);
    
#if UseAlpha
      if (Control.output>1 && myid==0)
        printf("Eg= %18.14f alpha/Pi= % 10.8f d_alpha/M_PI=% 9.1e\n",
               energy, alpha/M_PI, d_alpha/M_PI);
#if SearchGammaMin
      if (Control.output>1 && myid==0)
         printf(" : gamma/Pi = % 10.8f : d_beta=% 10.2e\n",
	           gamma/M_PI,d_gamma);
      if (energy > old_gamma_Energy) d_gamma = -0.5 * d_gamma; // go back
#else
    //  if (d_beta*(beta+d_beta) >= d_beta*(StopAlpha))
          d_gamma = 0;                            /* stop condition */
#endif
		 
      old_gamma_Energy = energy; // last value
     }

#if SearchBetaMin
      if (Control.output>1 && myid==0)
         printf(" : beta/Pi = % 10.8f : d_beta=% 10.2e\n",
	           beta/M_PI,d_beta);
      if (energy > old_beta_Energy) d_beta = -0.5 * d_beta; // go back
#else
    //  if (d_beta*(beta+d_beta) >= d_beta*(StopAlpha))
          d_beta = 0;                            /* stop condition */
#endif
		 
      old_beta_Energy = energy; // last value
     }

#if SearchAlphaMin
      if (energy > old_Energy) d_alpha = -0.5 * d_alpha; // go back
#else
//      if (d_alpha*(alpha+d_alpha) >= d_alpha*(StopAlpha))
          d_alpha = 0;                            /* stop condition */
#endif
      old_Energy = energy; // last value
      if (energy<minE && iterations<MAX_NUM_ITERATIONS) { // Mar08
        minE=energy; a_minE=alpha;
      } 
    }
    
    // alpha=a_minE; energy=minE; // ???
    if (Control.output>-1 && myid==0)
        printf("E(alpha)_lowest: Eg = %12.8f : alpha/Pi = % 10.8f\n",
               minE, a_minE/M_PI); // Mar08
    if (Control.output>-1 && myid==0)
        printf("E(alpha)_last:   Eg = %12.8f : alpha/Pi = % 10.8f\n",
               energy, alpha/M_PI); // Mar08
#if SearchGammaMin
    if (Control.output>-1 && myid==0)
        printf("E(gamma)_lowest: Eg = %12.8f : gamma/Pi = % 12.8f\n",
               energy, gamma/M_PI);
#endif
#if SearchBetaMin
    if (Control.output>-1 && myid==0)
        printf("E(beta)_lowest: Eg = %12.8f : beta/Pi = % 12.8f\n",
               energy, beta/M_PI);
#endif

#endif

    if (iterations >= MAX_NUM_ITERATIONS) continue;
      // break;

    time(&t0);

    if (solutionStrategy==1) {
      
     // (a) LU decomp. bra state.

      findBraState(x, bra_x);
      iterations2=100; // ???

    } else {
      
      // (b) direct iterations for bra states.

      status=directIterationBraState((int) MAX_NUM_ITERATIONS, 
	     NumberConfigurations, (double) TOLERANCE, 
	     iterations2, x, bra_x); 
    }

    time(&t1);

    // (c) the (sub)lattice magnetisation.
      
    stagMag=findStagMag(x,bra_x,subLatticeMag);
	
    if (Control.output>0 && myid==0) // print out later?
        printf(" M= %18.14f it= %4d  Mg.i100t= %6.2fm\n",
           stagMag, iterations2, (t1-t0)/.6/iterations2);
    
	
    // now find the generalised expectations.
	
    if (Control.greatExpectations) findGenExp(x,bra_x,g_exp,alpha,beta,gamma);

    // PRINT OUT INFORMATION.
    
    // we need at least two predecessing steps
    if ( fabs(tdelta-start) <= fabs(1.1*increment) ) {
      dEnergy  = ddEnergy  = 0;
      dStagMag = ddStagMag = 0;
    } else {
      dEnergy  = (energy-Energy1)/(increment);
      ddEnergy = (energy-2*Energy1+Energy2)/(increment*increment);
      dStagMag = (stagMag-stagMag1)/(increment);
      ddStagMag= (stagMag-2*stagMag1+stagMag2)/(increment*increment);
      if (Control.output > 0 && myid == 0) {
        printf("Computational 1st Derivative Eg wrt delta = % f\n",dEnergy);
        printf("Computational 2nd Derivative Eg wrt delta = % f\n",ddEnergy);
        printf("Computational 1st Derivative  M wrt delta = % f\n",dStagMag);
        printf("Computational 2nd Derivative  M wrt delta = % f\n",ddStagMag);
      }
    }    
    // = -d^2 E_g/ d Delta^2 via computational evaluation only.
    
    Energy2=Energy1;      Energy1=energy;   // save older values
    stagMag2=stagMag1;    stagMag1=stagMag;

    if (Control.output > -1 && myid == 0) {
      printf("Delta = %6.4f : Eg = %12.8f : M = %12.8f : -dEg = %12.8f : Heat = %12.8f\n"
             "  x[1]=%12.9f : iterations = %4d\n"
             "bra[1]=%12.9f : iterations = %4d\n",
	     sign*tdelta, energy, stagMag, -dEnergy, ddEnergy,
             x[1], iterations, bra_x[1], iterations2);
      if(UnitCell.number>1) {
	for(k=1;k<=UnitCell.number;k++)
	  printf("Mag.[%d] = %12.8f\n", k, subLatticeMag[k]);
    }
#if UseAlpha
  printf("Alpha=%16.12f\n ",alpha/M_PI);
#endif
#if SearchBetaMin
  printf("Beta=%16.12f\n",beta/M_PI);
#endif
#if SearchGammaMin
  printf("Gamma=%16.12f\n",gamma/M_PI);
#endif
	// printf("\n");
      
  if (Control.greatExpectations>0) {
    for (k=1;k<=Params.nExpect;k++)
      printf("Generalised Expectation Value [%d] = %12.8f\n", k, g_exp[k]);
    }
    printf("Lattice Magnetisation Triangular AF=%12.8f\n",g_exp[1]+g_exp[2]); 
    if (Control.output > 0) {
      int i, j0, j1;
      for(j0=j1=0,i=1;i<=NumberConfigurations;i++) {
        if (x[i]<0) j0++; if(bra_x[i]<0) j1++;
      }
      printf("\nnumber negativ coefficients (x,bra_x) = %d %d"
               "  NumConf= %d\n", j0, j1, NumberConfigurations);
      if (j0<100 && Control.output>1) { // lot of output
          for(j0=0,i=1;i<=NumberConfigurations;i++) {
            if (x[i]<0) {
              if(j0<1000) printf(" %d",i);
              if(j0==1000) printf("\n");
              j0++;
            }
          }
          printf("\n");
          for(j0=0,i=1;i<=NumberConfigurations;i++) {
            if (bra_x[i]<0) {
              if(j0<1000) printf(" %d",i);
              if(j0==1000) printf("\n");
              j0++;
            }
          }
          printf("\n");
        }
      }
      if (Control.output > 1) {
        int i; double sum=0,sumb=0,prod=0;
        for(i=1;i<=NumberConfigurations;i++) {
          if (i<Control.output*20)
            printf(" i x[i] bra[i] %4d % .9f % .9f\n",
                     i,x[i],bra_x[i]);
          if (i==Control.output*20) printf("...\n");
          sum+=x[i]*x[i]; sumb+=bra_x[i]*bra_x[i]; prod+=x[i]*bra_x[i];
        }
        printf(" x.x bra.bra x.bra  % .9f % .9f % .9f\n",sum,sumb,prod);
      }
    }

    if (myid == 0) {
      fprintf(fp2,"%16.12f %16.12f %16.12f %16.12f %16.12f %16.12f %d %d",
	sign*tdelta,energy,stagMag,-dEnergy,ddEnergy,x[1],iterations,iterations2);
#if UseAlpha
  fprintf(fp2,"%16.12f ",alpha/M_PI);
#endif
#if SearchBetaMin
  fprintf(fp2,"%16.12f ",beta/M_PI);
#endif
#if SearchGammaMin
  fprintf(fp2,"%16.12f ",gamma/M_PI);
#endif
      if(UnitCell.number>1) {
	for(k=1;k<=UnitCell.number;k++)
	  fprintf(fp2,"%16.12f ",subLatticeMag[k]);
      }
  	 if (Control.greatExpectations) {
	  for (k=1;k<=Params.nExpect;k++)
	     fprintf(fp2,"%16.12f ", g_exp[k]);
      }
	  fprintf(fp2,"%16.12f ",g_exp[1]+g_exp[2]); // for triangular AD lattice mag only
      fprintf(fp2,"\n");
      fflush(fp2);
    }
  }
  if (Control.output > -1 && myid == 0) {
    printf("============================================================\n\n");
    printf("SOLVE CCM EQUATIONS: SUCCESSFULLY COMPLETED!\n\n");
    printf("============================================================\n\n");
  }

  // fclose(terms_fp);    terms_fp=NULL;
  if (myid==0) fclose(fp2);

 // ferase(fp_exp);

  // stat=remove(gs_name);
  // stat=remove(input_name);  
  // stat=remove(configurations);

}

/* called by solve() and directIteration(only_printout)
 * new: x[0]=1 now (Sep05)
 * 
 */
double findEnergy(double *x) 
{

  int i, bonds;
  int indices[7]={0};
  double dummy;
  double sum;

  sum=0.0; // nio=7;
  
  term_rewind(0);

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    for(i=1;;i++) {
      
      term_get(0,indices,&dummy);  // get gse_terms
      
      if (indices[0]<0 || indices[0] > NumberConfigurations) break;
      
      // Joerg Sep05:
      //   according to gse_terms only indices[2..3] can be nonzero
      // gse_term{3}      : (i2,i3) = two single-site configs
      // gse_term{2,4,5,7}: (i2, 0) = two-site neighbour config
      // gse_term{1,6,8}  : ( 0, 0)
      // squareLSUB10 -> B1: (0,1,0,0,0,0,-0.5), B2: (0,27,1,0,0,0,-1.0)
      sum += dummy * x[indices[2]]
                   * x[indices[3]] * delta[bonds][indices[1]];
    }

  }
  return sum;
}
 
// in/out: oldX - startvector + iterationvector
// in(out: dx   - difference oldX(out) - oldX(in)  (only if precondition)
//                used to modify start vector (speed up?) -> removed
int directIteration(int maximumNumIterations, int numConfigs, 
		    double tolerance, int &iterations, double *oldX)
{  
  int i, j, bonds, max_err_j[3];
  int indices[7]={0};
  double dummy;
  double xi2, xi3, xi4, xi5;
  double *lhs, *rhs,    /* Joerg: left-hand side, right hand side ? */ 
         min_error, error, mom, err;
  double max_err[3]; // store 3 biggest changes
//  double tempSum1, sum1, tempSum2, sum2;
//  double *tempLHS,  *tempRHS;
//  double *tempLHS2, *tempRHS2;

  error = 0.0; // JS Nov05
  min_error = 10.0; // JS Apr06, to stop divergence problems LSUB14 chain
  mom=momentum;     // default is 0.7 set in Constants.h 

  if (myid==0)
    printf("Direct iteration ket state (tolerance=%g, momentum=%g):\n",
            tolerance, momentum);

  lhs= dvector(0,numConfigs+5);
  rhs= dvector(0,numConfigs+5);

  for(i=1;i<=numConfigs;i++) {
    lhs[i]=0.0;
    rhs[i]=0.0;
  }
  
  /* Starting values */
  if (Control.output > 0 && myid == 0) 
      printf(" iteration= %4d  energy= %16.10f  oldX[1]= %16.10f\n",
	     0, findEnergy(oldX), oldX[1]);

  for(iterations=1; iterations<=maximumNumIterations; iterations++) {
    
    term_rewind(1);

    /* calculate rhs and lhs from input-file */
    for(bonds=1; bonds<=Params.bonds; bonds++) {
      
      /* why loop over bonds? inner loop does not depend on bonds? */
      for(i=1;;i++) {
	
        term_get(1,indices,&dummy);
	
	if (indices[0] < 0 || indices[0] > NumberConfigurations)
	  break;
	
        /* now: oldX[0]=1.0, Joerg Sep05 */
        xi2=oldX[indices[2]]; /* pointing to a configuration index or 0 */
        xi3=oldX[indices[3]];
        xi4=oldX[indices[4]];
        xi5=oldX[indices[5]]; /* 4 indices only for term34 */
    
        /* ToDo: indices are sorted, so that zeros are at end (indices[5]) */
        /* Joerg: we look for terms where indices[2]==indices[0] and rest 0 */
        if (indices[3]==0 && indices[2]==indices[0])
          /* Joerg Nov05: diagonal term? called rare */
          lhs[indices[0]]-=dummy*delta[bonds][indices[1]];
        else { /* called often */
          /* Joerg Nov05: offdiagonal terms? */
          rhs[indices[0]]+=dummy*xi2*xi3*xi4*xi5
                                *delta[bonds][indices[1]];
        }
      }
    }

    error=0.0;

    // only DEBUG (myid1=1 myid2=2 ... myid0=npe-1)
    for(j=((myid==0)?npe:myid);j<=numConfigs;j+=npe) {  // parallel 1...
      /* old: using tolerance=1e-12 here is not a good idea!? */
      if (fabs(lhs[j]) < 1.0e-9) {  // usually -0.5,-1.0,...
	printf("Error in direct iterations: no left-hand side...\n"
          "it= %4d  cfg= %4d  lhs= % .4g\n"
	  "Exiting!!\n\n", iterations, j, lhs[j]);
        exit(1);
      }
    }
#if (PARALLEL_FLAG> 0) 
    /* MPI_Barrier(MPI_COMM_WORLD); dont needed!? */
    /* new faster code by JS */
    /* only every npe-th element is nonzero, summarize */
    // we use lhs temporary to sum over threads
    for(j=1;j<=numConfigs;j++) {   /* [0] is not used */
      if ((j%npe) == myid) lhs[j]=rhs[j]/lhs[j]; else lhs[j]=0.0;
    }
    // store the sum of lhs to rhs
    MPI_Allreduce(&lhs[1], &rhs[1], numConfigs,
                  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    for(j=1;j<=numConfigs;j++) lhs[j]=1.0; // parallel-to-single adaption
    // now we can use single-task algorithm to go further
#endif

    max_err[0]=max_err[1]=max_err[2]=0; 
    max_err_j[0]=max_err_j[1]=max_err_j[2]=0;
    for(j=1;j<=numConfigs;j++) {
      error+=fabs(err=oldX[j]-rhs[j]/lhs[j]);
      // store 3 biggest changes
      if (fabs(err)>fabs(max_err[0])) {
         max_err[2]=max_err[1]; max_err_j[2]=max_err_j[1];
         max_err[1]=max_err[0]; max_err_j[1]=max_err_j[0];
         max_err[0]=err; max_err_j[0]=j; 
      } else
      if (fabs(err)>fabs(max_err[1])) {
         max_err[2]=max_err[1]; max_err_j[2]=max_err_j[1];
         max_err[1]=err; max_err_j[1]=j; 
      } else
      if (fabs(err)>fabs(max_err[2])) {
         max_err[2]=err; max_err_j[2]=j; 
      }
      // ToDo: get j of max(err)
    }
    // error = fabs(max_err[0]); // other |X|=distance definition
#if 0  // can generate lot of iterations
    if (error > 2*min_error) {   // should go down allways
      mom = 1.0 - 0.5*(1.0-mom);
      min_error = error;
      if (myid==0)
        printf(" It=%d: reduce 1-momentum to %g to ensure convergence\n",
             iterations, 1-mom);
    }
#endif
    // max_X_j=0; max_X=0.0; // only DEBUG
    for(j=1;j<=numConfigs;j++) {
      oldX[j]=mom*oldX[j]+(1.0-mom)*rhs[j]/lhs[j]; // some parallel overhead
      lhs[j]=0.0;
      rhs[j]=0.0;
      // ToDo: outside It.: if (fabs(oldX[j])>fabs(max_X)) {max_X=oldX[j]; max_X_j=j;}
    }
    
    /* calling findEnergy() here is a vast of time ??? (Joerg) */
    if (Control.output > 0 && myid == 0 && ((iterations % 10)==0 
      || Control.output> 2 || error > 0.0625 || min_error == error)) {
      // err= sum(fabs(dX[i]),i) (i,dX[i])_for_max(fabs(dX[i])
      printf(" %4d energy= %16.10f err= %9.2e %4d %9.2e %4d %9.2e\n",
	     iterations, findEnergy(oldX), error,
             max_err_j[0], max_err[0], max_err_j[1], max_err[1]);
    }

    if (isnan(error)) {   /* break if +-Inf or NaN */
       printf(" iteration = %4d : energy = %16.10f : "
              "error = %14.12f\nError directIteration: NaN-BREAK\n",
              iterations, findEnergy(oldX), error);
       iterations=maximumNumIterations; break;
    }

    if (min_error > error) min_error = error;
    // min_error=fabs(max_err[0]); mom=momentum;
    if (error < tolerance) break; // convergence
    if (error > 1e6) break;       // divergence

  }

  if (iterations>=maximumNumIterations) {
    // if (!myid)
    printf("directIteration Error: iterations>=maximumNumIterations\n");
#if Parallel_Flag
    MPI_Abort(MPI_COMM_WORLD,0);
#else
    exit(0);
#endif
  }

  free_dvector(lhs,0,numConfigs+5);
  free_dvector(rhs,0,numConfigs+5);

  return iterations;
  
}

// in: x[NumConf]
// out: fvec[NumConf], fjac[NumConf][NumConf]
// called by mnewt()=NewtonRawson in num_methods.C
void usrfun(int n, double *fvec, double **fjac, double *x, int iterations)
{  
  int i, k, bonds;
  int indices[7]={0};
  double dummy;
  double xi2, xi3, xi4, xi5, multiplier;
  static int warn1=0;
  
  term_rewind(1);
  
  if (sizeof(int *)<=4 && NumberConfigurations>=16384 && myid==0 && warn1==0){
    /* may be the 2D-array is to big for a 32bit-system (nC>=(2GB/8)^.5) */
    printf("Warning usrfun: array may be to big for a 32bit-system\n");
  }
  for(i=1;i<=NumberConfigurations;i++) {    
    fvec[i]=0.0;
    for(k=1;k<=NumberConfigurations;k++) 
      fjac[i][k]=0.0;  /* may fail if nC*nC*double>2GB */
  }  
 
  i=0;
  for(bonds=1;bonds<=Params.bonds;bonds++) {

    for(;;i++) {
      
      term_get(1,indices,&dummy);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations)
	break;
      
      x[0]=1.0; /* obsolete??? Joerg Sep05 */
      xi2=x[indices[2]];
      xi3=x[indices[3]];
      xi4=x[indices[4]];
      xi5=x[indices[5]]; 

      multiplier = delta[bonds][indices[1]]*dummy;
      
      fvec[indices[0]] += xi2*xi3*xi4*xi5*multiplier;
      
      if (indices[2] > 0) fjac[indices[0]][indices[2]] += xi3*xi4*xi5*multiplier;
      if (indices[3] > 0) fjac[indices[0]][indices[3]] += xi2*xi4*xi5*multiplier;
      if (indices[4] > 0) fjac[indices[0]][indices[4]] += xi2*xi3*xi5*multiplier;
      if (indices[5] > 0) fjac[indices[0]][indices[5]] += xi2*xi3*xi4*multiplier;
       
    }
  }

  if (Control.output > 10 && myid == 0) {
    printf("usrfun (read %d):\n",i);
    for (i=1; i<=NumberConfigurations; i++) {
      if (i==6 && NumberConfigurations>10) printf (" ...\n");
      if (i>5 && i<NumberConfigurations-4) continue;
      for (k=1; k<=NumberConfigurations; k++) {
        if (k<=5 || k>NumberConfigurations-5) printf(" % 10.8f",fjac[i][k]);
        if (k==6 && NumberConfigurations >10) printf (" ...");
      }
      printf(" : % 10.8f \n",fvec[i]);
    }
  }
  
}

/*****************************************************
 *                                                   *
 * Obtain bra state (LU decomp.)                     *
 *                                                   *
 *****************************************************
 *                                                   *
 * x_I        -- ket-state correlatin coefficient.   *
 * \tilde x_I -- bra-state correlatin coefficient.   *
 *                                                   *
 * H = e_0 + sum_I \tilde x_I F_I                    *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d (\tilde x_I) =>                           *
 *                                                   *
 * F_I = 0    <= ket-state equations                 *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d x_I =>                                    *
 *                                                   *
 * d(e_o)/d(x_I)+sum_J \tilde x_J (d(F_J)/d(x_I)) =0 * 
 *            <= bra-state equations                 *
 *                                                   *
 * use LU decomposition to solve the equations:      *
 *                                                   *
 * F(I,J) (\tilde X(J)) = - e_o(I)                   *
 *                                                   *
 *****************************************************/

void findBraState(double *x, double *bra_x)
{

  /* Determines the bra-state correlation coefficients
     using LU decomposition.
     */

  int      i;
//  int      status;
  void     lubksb(double **a, int n, int *indx, double b[]);
  void     ludcmp(double **a, int n, int *indx, double *d);
  int      *ind;
  double   d;
  double   **bra_eqn;
  // int      length;
  // int      iterations;

  for(i=1;i<=NumberConfigurations;i++) 
    bra_x[i]=0.0; 

  ind=ivector(0,NumberConfigurations+1);
  bra_eqn=dmatrix(0,NumberConfigurations+1,0,NumberConfigurations+1); 

  findBraStateRHS(bra_x,x); // bra_x (-=) terms(x)
  
  // find sum_J \tilde x_J (d(F_J)/d(x_I)) == F(I,J)
  
  findBraEquations(bra_eqn, x);
  
  // LU decomp Numerical Recipes to solve: F(I,J) (\tilde X(J)) = - e_o(I)
  
  ludcmp(bra_eqn,NumberConfigurations,ind,&d);
  lubksb(bra_eqn,NumberConfigurations,ind,bra_x);
  
  free_ivector(ind,0,NumberConfigurations+1);
  free_dmatrix(bra_eqn,0,NumberConfigurations+1,0,NumberConfigurations+1);

}
  

/*****************************************************
 *                                                   *
 * Obtain bra state (direct iteration)               *
 *                                                   *
 *****************************************************
 *                                                   *
 * x_I        -- ket-state correlatin coefficient.   *
 * \tilde x_I -- bra-state correlatin coefficient.   *
 *                                                   *
 * H = e_0 + sum_I \tilde x_I F_I                    *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d (\tilde x_I) =>                           *
 *                                                   *
 * F_I = 0    <= ket-state equations                 *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d x_I =>                                    *
 *                                                   *
 * d(e_o)/d(x_I)+sum_J \tilde x_J (d(F_J)/d(x_I)) =0 * 
 *            <= bra-state equations                 *
 *                                                   *
 * Use direct iteration to solve:                    *
 *                                                   *
 *    (\tilde X(I)) = { - e_o(I) - \sum_{J != I}     *
 *                F(I,J) (\tilde X(J))}/F(I,I)       *
 *                                                   *
 * for the \tilde X(I)'s.                            *
 *                                                   *
 *****************************************************/

int directIterationBraState(int maximumNumIterations, int numConfigs, 
			    double tolerance, int &iterations, 
			    double *oldX, double *oldBraX)
{

  double error, min_error, mom;
  
  int i, j, bonds;
  int indices[7]={0};
  double dummy[1];
  double xi2, xi3, xi4, xi5,multiplier;
//  double tempSum1, sum1,tempSum2, sum2;
  double subLatticeMag[MAX_UNIT_CELL+3];

  double *lhs, *rhs;
#if (PARALLEL_FLAG> 0)
  double *tempLHS, *tempRHS;
  tempLHS=dvector(0,numConfigs-1);
  tempRHS=dvector(0,numConfigs-1);
#endif
  
  error = 0.0;
  min_error= 10.0;
  mom = momentum;
  // mom = 0.7;  // 0.25
  if (myid == 0) printf("Direct iteration bra state (mom=%g):\n",mom);

  lhs=dvector(0,numConfigs+5);
  rhs=dvector(0,numConfigs+5);

  for(i=1;i<=numConfigs;i++) {
    lhs[i]=0.0;
    rhs[i]=0.0;
  }

  if (Control.output > 0 && myid == 0) 
    printf(" %4d  mag= %12.10f  braX[1]= %12.10f\n",
      0,findStagMag(oldX,oldBraX,subLatticeMag),oldBraX[1]);

  for(iterations=1;iterations<=maximumNumIterations;iterations++) {
        
#if (PARALLEL_FLAG == 0) 
    findBraStateRHS(rhs,oldX); // rhs (-=) terms(oldX)
#endif

    term_rewind(1);

    for(bonds=1;bonds<=Params.bonds;bonds++) {
      
      for(i=1;;i++) {
	
        term_get(1,indices,dummy);
	
	if (indices[0] < 0 || indices[0] > NumberConfigurations)
	  break;

        multiplier=delta[bonds][indices[1]];

	
	if (indices[2]==0) { // all indices are equal to 0
	  
	  // rhs[indices[0]]+=dummy[0]*multiplier;
	  
	} else {
	  
          oldX[0]=1.0;
	  xi2=oldX[indices[2]];
	  xi3=oldX[indices[3]];
	  xi4=oldX[indices[4]];
	  xi5=oldX[indices[5]];  /* only for term34 */

         if (indices[5]>0) {  /* only term34 */
              rhs[indices[5]]-=dummy[0]*xi2*xi3*xi4*multiplier*
                oldBraX[indices[0]];
          }

          if (indices[4]>0) {
              rhs[indices[4]]-=dummy[0]*xi2*xi3*xi5*multiplier*
                oldBraX[indices[0]];
          }

          if (indices[3]>0) {
              rhs[indices[3]]-=dummy[0]*xi2*xi4*xi5*multiplier*
                oldBraX[indices[0]];
          }

	  if (indices[2]>0) {
	    if (indices[2]==indices[0]  && indices[3] ==0) // [4..5] == 0 too
	      lhs[indices[0]]+=dummy[0]*xi3*xi4*xi5*multiplier;
	    else 
	      rhs[indices[2]]-=dummy[0]*xi3*xi4*xi5*multiplier*
		oldBraX[indices[0]];
	  }
	}
      }
    }
      
#if (PARALLEL_FLAG> 0)
    /* ToDo JS06: speedup if rhs/lhs or only error[npe] summed,
     *    changing oldX can be distributed over threads */
    MPI_Allreduce(lhs+1, tempLHS, numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    MPI_Allreduce(rhs+1, tempRHS, numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    for(i=1;i<=numConfigs;i++) {
      lhs[i]=tempLHS[i-1];
      rhs[i]=tempRHS[i-1];
    }
    findBraStateRHS(rhs,oldX);  // rhs (-=) terms(oldX)
#endif

    error=0.0;
    for(j=1;j<=numConfigs;j++) {
      if (fabs(lhs[j]) < 1e-9) {
	printf("Error in direct iterations bra state: no left-hand side...\n"
	       "Exiting!!\n\n");
	break;
      }
      error+=fabs(oldBraX[j]-rhs[j]/lhs[j]);
    }
    if (j<=numConfigs) break;
#if 0  // can generate lot of iterations
    if (error>8*min_error) { // New JS Apr06, 2* => 8* Mar08, ToDo: make bigger?
       mom = 1.0 - 0.5*(1.0 - mom);  /* 1-0.3 -> 1-0.15 -> 1-0.075 -> ... */
       min_error = error;
       if (!myid)
         printf(" It=%d: reduce 1-momentum to %g to ensure convergence\n",
                iterations, 1.0-mom);
       if (1.0-mom<1e-6) break;
    }
#endif
    for(j=1;j<=numConfigs;j++) {
      oldBraX[j]=mom*oldBraX[j]+(1.0-mom)*rhs[j]/lhs[j];
      lhs[j]=0.0;
      rhs[j]=0.0;
    }

    if (Control.output > 0 && myid == 0
       && ((iterations % 20)==0 || Control.output > 2)) 
      printf(" %4d  mag= %12.10f err= %9.2e %9.2e\n",
	iterations,findStagMag(oldX,oldBraX,subLatticeMag),error,error-min_error);

    if (error < min_error) min_error = error;
    if (error < tolerance) break;

  }

  if (iterations>=maximumNumIterations) {
    printf("directIterationBraState Error: i>=maximumNumIterations\n");
    ERR(0);
  }

  free_dvector(lhs,0,numConfigs+5);
  free_dvector(rhs,0,numConfigs+5);
#if (PARALLEL_FLAG> 0)
  free_dvector(tempLHS,0,numConfigs-1);
  free_dvector(tempRHS,0,numConfigs-1);    
#endif

  return iterations;
}

// input:  rhs, x (global: terms, Params.bonds, NumberConfigurations)
// output: rhs
// --- called from each thread ---
void findBraStateRHS(double *rhs, double *x) 
{
  int i, bonds;
  int indices[7]={0};
  double dummy[1];
  double xi2, xi3, xi4, xi5, multiplier;

  term_rewind(0);

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    for(i=1;;i++) {
      
      term_get(0,indices,dummy);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations) break;
      
      multiplier=delta[bonds][indices[1]];

      // ToDo: how often? can be removed?
      if ((indices[2]==0 && indices[3]==0
        && indices[4]==0 && indices[5]==0)) continue;
	
        x[0]=1;
        xi2=x[indices[2]];
	xi3=x[indices[3]];
	xi4=x[indices[4]];
	xi5=x[indices[5]];
	
	if (indices[2] > 0) // ToDo: faster if using dummy rhs[0] ??? 
	  rhs[indices[2]]-= dummy[0]*xi3*xi4*xi5*multiplier;
	
	if (indices[3] > 0) 
          rhs[indices[3]]-= dummy[0]*xi2*xi4*xi5*multiplier;
	
	if (indices[4] > 0)
	  rhs[indices[4]]-= dummy[0]*xi2*xi3*xi5*multiplier;
	
	if (indices[5] > 0)
	  rhs[indices[5]]-= dummy[0]*xi2*xi3*xi4*multiplier;
    }
  }
}

void findBraEquations(double **m, double *x) 
{
  /* FIND BRA EQUATIONS: finds the Jacobian in order to determine 
   the bra state equations.
   */

  // find: sum_J \tilde x_J (d(F_J)/d(x_I))
  
  int i,k,bonds;
  int indices[7]={0};
  double dummy[1];
  double xi2, xi3, xi4, xi5, multiplier;
  
  term_rewind(1);
  
  for(i=1;i<=NumberConfigurations;i++)    
    for(k=1;k<=NumberConfigurations;k++) 
      m[i][k]=0.0;

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    
    for(i=1;;i++) {
      
      term_get(1,indices,dummy);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations)
	break;
      
      multiplier=delta[bonds][indices[1]];

      if (!(indices[2]==0 && indices[3]==0 && indices[4]==0 && 
	    indices[5]==0)) {
	
        x[0]=1.0;
	xi2=x[indices[2]];
	xi3=x[indices[3]];
	xi4=x[indices[4]];
	xi5=x[indices[5]];
	
	if (indices[2] > 0)
	  m[indices[2]][indices[0]]+= dummy[0]*xi3*xi4*xi5*multiplier;
	
	if (indices[3] > 0)
	  m[indices[3]][indices[0]]+= dummy[0]*xi2*xi4*xi5*multiplier;
	
	if (indices[4] > 0)
	  m[indices[4]][indices[0]]+= dummy[0]*xi2*xi3*xi5*multiplier;
	
	if (indices[5] > 0)
	  m[indices[5]][indices[0]]+= dummy[0]*xi2*xi3*xi4*multiplier;
      }
    }
  }
}

/*****************************************************
 *                                                   *
 * Name: Staggered Magnetisation Routine.            *
 *                                                   *
 * Determines the lattice magnetisation once bra-    *
 * and ket-state equations have been solved.         *
 *****************************************************/

double findStagMag(double *x, double *bra_x, double *subLatticeMag)
{

  /* FIND STAG MAG: obtains the staggered magnetisation.
     */

  int      i, j;
  double   stagMag;
  int      length;
  int      (*configuration)[5];
  int      subTerm[MAX_UNIT_CELL+3];

  configuration = (int (*)[5]) malloc( (MAX_LSUBM+1)*5*sizeof(int) );

  // first find the \sum_i < \tilde \Psi | s_i^z | \Psi > bit.

  stagMag=0.0; 
  
  for(i=1;i<=UnitCell.number;i++) {
    subLatticeMag[i] = 0.5*UnitCell.s2[i];
    stagMag         += 0.5*UnitCell.s2[i];
  }
  
  stagMag=stagMag/((double) UnitCell.number);

  for(i=1;i<=NumberConfigurations;i++) {

    length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
				     configuration);

    stagMag -= length*bra_x[i]*x[i];    
    sublattice_term(length,configuration,subTerm);
    
    for(j=1;j<=UnitCell.number;j++) 
      subLatticeMag[j] -= subTerm[j] * bra_x[i] * x[i]
                        * UnitCell.number / LatticeSymmetries.number;
  }

  free(configuration);
  return stagMag;
}

void sublattice_term(int length, int (*configuration)[5], int *subTerm)
{
  
  /* 
     This routine finds contribution to the sublattice magnetisation
     from a given cluster, namely:
     
     F(m) . S(+)(m) 
  */
  int i, j;
  static int tempConfiguration[MAX_LSUBM+1][5];
  
  for(i=1;i<=UnitCell.number;i++) subTerm[i]=0;

  // find symmetry related clusters
  for(j=1;j<=LatticeSymmetries.number;j++) { 
    /* ToDo: configuration modified ?? */
    symmetryGroup(j,length,configuration,tempConfiguration);
    for(i=1;i<=length;i++) { subTerm[tempConfiguration[i][4]]++; }
  } 

  return;
}

// new code to find the generalised expectation values.

void findGenExp(double *x, double *bra_x, double *g_exp, double alpha, double beta, double gamma)
{

  /* FIND GENERALISED EXPECTATION VALUES: obtains the staggered magnetisation.
     */

  int      i,j,index,term;
  int      (*configuration)[5];
  int      expect,error;
  int	   indices[7]={0};
  double   dummy[1],multiplier;
  double   xi2, xi3, xi4, xi5,bx;

  configuration = (int (*)[5]) malloc( (MAX_LSUBM+1)*5*sizeof(int) );
    
  for(i=0;i<=Params.nExpect;i++) {
    g_exp[i]=0.0;
  }

  term_rewind(2); // term_open(2,"r"); Jun08

  for(expect=1;expect<=Params.nExpect;expect++) {
    for (j=1;j<=2;j++) {
      for(i=1;;i++) {

        error=term_get(2,indices,dummy); /* .exp_terms */
        if (indices[0] < 0 || indices[0] > NumberConfigurations) break;
        index=indices[0];
        // the index for the bra-state correlation coefficient.
	// if this is zero then it's one of the "gs" like terms.
	
	term=indices[1];
	bx=bra_x[index];
	if (index < 1) bx=1.0;
	
	xi2=x[indices[2]];
	xi3=x[indices[3]];
	xi4=x[indices[4]];
	xi5=x[indices[5]];

        multiplier=1.0;

#if UseExpAlpha
	multiplier=usr_exp_angle(expect,term,alpha,beta,gamma);
#endif
			
	if (Control.output > 3 && myid ==0) 
	  printf("expect=%d : config=%d : term =%d : (%d,%d,%d,%d) : mult1=%10.6f : mult2=%10.6f : g_exp=%10.6f\n",
	         expect,indices[0],indices[1],indices[2],indices[3],indices[4],indices[5],
		 dummy[0],multiplier,g_exp[expect]);

        g_exp[expect] += bx*xi2*xi3*xi4*xi5*dummy[0]*multiplier;
			
      }
    }
  }

  // fclose(fp_exp); Jun08
  free(configuration);
  return;
}

